package simulator;

import java.util.*;
import java.io.*;
import javax.media.j3d.*;

import util.*;

public class Simulation
{
    public static Hashtable sysModels = new Hashtable();
    public static Hashtable cellModels = new Hashtable();
    public static Hashtable parsers = new Hashtable();
    
    private Properties configuration = new Properties();
    private BufferedReader cellProgramReader;
    private SimFileTokenizer simFileTokenizer;
    private BufferedWriter outputWriter;
    private BufferedWriter statsWriter;
    private BufferedWriter errorWriter;
    
    private String systemModelName;
    private String systemModelOptions;
    private String cellModelName;
    private String cellModelOptions;
    private String parserModelName;
    private String parserModelOptions;
    private String title;
    private SimToken nextSimToken;
    
    private SystemModel sys;
    private CellProgram cellProgram;
    
    private int step = 0;
    
    public Simulation (SimulationParameters params)
    {
	/*
	** Read sim file
	*/
	
	try {
	    String simFileName = params.getSimFile ();
	    title = simFileName;
	    BufferedReader simFileStream = new BufferedReader (new FileReader (simFileName));
	    simFileTokenizer = new SimFileTokenizer (simFileStream);
	} catch (FileNotFoundException e) {
	    System.err.println ("Unable to find sim file " + params.getSimFile () + ": " + e.toString ());
	    return;
	} catch (IOException e) {
	    System.err.println ("Unable to open sim file: "+ e.toString());
	    return;
	}

	try {
	    boolean sysModelRead = false, cellModelRead = false, parserModelRead = false;
	    
	    nextSimToken = simFileTokenizer.nextSimToken ();

	    while ((nextSimToken.getType() == SimToken.SIM_CONFIG)
		   && (!sysModelRead || !cellModelRead || !parserModelRead))
		{
		    if (nextSimToken.getType() == SimToken.SIM_CONFIG)
			{
			    if (nextSimToken.getConfigParam().equals("system"))
				{
				    configuration.setProperty("system", nextSimToken.getConfigValue());
				    sysModelRead = true;
				}
			    else if (nextSimToken.getConfigParam().equals("cell"))
				{
				    configuration.setProperty("cell", nextSimToken.getConfigValue());
				    cellModelRead = true;
				}
			    else if(nextSimToken.getConfigParam().equals("parser"))
				{
				    configuration.setProperty("parser", nextSimToken.getConfigValue());
				    parserModelRead = true;
				}
			}
		    nextSimToken = simFileTokenizer.nextSimToken ();
		}
	    
	    if (!(sysModelRead && cellModelRead && parserModelRead))
		{
		    errorWriter.write("Incomplete specifications in sim file");
		    
		    if(!sysModelRead)
			errorWriter.write("System Model not specified");
		    else if(!cellModelRead)
			errorWriter.write("Cell Model not specified");
		    else
			errorWriter.write("Parser Model not specified");
		}
	} catch(IOException e) {
	    System.err.println("Io Exception ocurred reading sim file: " + e.toString());
	    System.exit(0);
	}

	systemModelName = new String (configuration.getProperty ("system"));
	systemModelOptions = new String ("");
	cellModelName = new String (configuration.getProperty ("cell"));
	cellModelOptions = new String ("");
	parserModelName = new String (configuration.getProperty("parser"));
	parserModelOptions = new String ("");

	/*
	** Create output file
	*/

	String outFileName = params.getOutFile ();

	if (outFileName != null) {
	    try {
		outputWriter = new BufferedWriter (new FileWriter (outFileName));
	    } catch (FileNotFoundException e) {
		System.err.println ("Unable to find output file " + outFileName + ": " + e.toString ());
		return;
	    } catch (IOException e) {
		System.err.println ("Unable to open output file " + outFileName + ": "+  e.toString ());
		return;
	    }
	} else {
	    outputWriter = new BufferedWriter (new PrintWriter (System.out));
	}

	String statsFileName = params.getStatsFile ();

	if (statsFileName != null) {
	    try {
		statsWriter = new BufferedWriter (new FileWriter (statsFileName));
	    } catch (FileNotFoundException e) {
		System.err.println ("Unable to find statsput file " + statsFileName + ": " + e.toString ());
		return;
	    } catch (IOException e) {
		System.err.println ("Unable to open statsput file " + statsFileName + ": "+  e.toString ());
		return;
	    }
	} else {
	    statsWriter = new BufferedWriter (new PrintWriter (System.out));
	}

	String errFileName = params.getErrFile ();

	if (errFileName != null) {
	    try {
		errorWriter = new BufferedWriter (new FileWriter (errFileName));
	    } catch (FileNotFoundException e) {
		System.err.println ("Unable to find errput file " + errFileName + ": " + e.toString ());
		return;
	    } catch (IOException e) {
		System.err.println ("Unable to open errput file " + errFileName + ": "+  e.toString ());
		return;
	    }
	} else {
	    errorWriter = new BufferedWriter (new PrintWriter (System.out));
	}

	Vector cellPrograms = params.getCellPrograms ();
	for (Enumeration e = cellPrograms.elements(); e.hasMoreElements();) {
	    String cellProgramName = (String) e.nextElement ();
	    BufferedReader cellProgramReader;

	    try {
		cellProgramReader = new BufferedReader (new FileReader (cellProgramName));
		try {
		    if (cellProgram == null) {
			cellProgram = createCellProgram (parserModelName, parserModelOptions, cellProgramReader);
		    } else {
			cellProgram.extend (createCellProgram (parserModelName, parserModelOptions, cellProgramReader));
		    }
		} catch (GenericParseException gpe) {
		    Assert.badBranch ("Parse error in " + cellProgramName + ": " + gpe.toString ());
		} catch (Exception exe) {
		    Assert.badBranch ("Error reading " + cellProgramName + ": " + exe.toString ());
		}
	    } catch (FileNotFoundException fnfe) {
		System.err.println ("Unable to find cell program file " + cellProgramName + ": " + fnfe.toString ());
	    } catch (IOException ioe) {
		System.err.println ("Error reading cell program file " + cellProgramName + ": " + ioe.toString ());
		return;
	    } 
	}
	
	try {
	    sys = createSystem (systemModelName, systemModelOptions);
	} catch (Exception e) {
	    System.err.println ("Cannot create system: " + e.toString ());
	}

	runInstructions ();
    }
    
    public BranchGroup getDisplayGraph()
    {
	return sys.getDisplayGraph();
    }
    
    public Hashtable getStatistics()
    {
	Vector argVector = new Vector();
	argVector.add(new String("Get Statistics"));
	Vector args = sys.provideService(argVector);
	
	/* Filter out and return the requested statistics */
	return (Hashtable) args.elementAt(0);
    }
    
    public CellProgram getCellProgram () {
	return cellProgram;
    }
    
    private void runInstructions ()
    {
	while (nextSimToken.getType() == nextSimToken.SIM_INSTRUCTION
	       && nextSimToken.getInstructionNumber() <= getStep())
	    {
		Vector args = new Vector();
		args.add (new String("Simulation Instruction"));
		args.add (nextSimToken.getInstruction());
		sys.provideService (args);
		try {
		    nextSimToken = simFileTokenizer.nextSimToken ();
		} catch(IOException e) {
		    System.err.println (e.toString ());
		}
	    }
    }
    
    public void nextStep()
    {
	step++;
	runInstructions ();
	sys.runCellProgramForOneStep ();
    }
    
    public int getStep ()
    {
	return step;
    }

    public String getTitle ()
    {
	return title + ": " + step;
    }
    
    public void writeOutputLine (String s) throws IOException
    {
	outputWriter.write(s,0,s.length());
	outputWriter.newLine();
    }
    
    public void writeStatsLine (String s) throws IOException
    {
	statsWriter.write(s,0,s.length());
	statsWriter.newLine();
    }
    
    public void writeErrorLine(String s) throws IOException
    {
	errorWriter.write(s,0,s.length());
	errorWriter.newLine();
    }
    
    public void finish()
    {
	try {
	    outputWriter.close();
	    statsWriter.close();
	    errorWriter.close();
	} catch(IOException e) {
	    System.err.println ("Error closing files: " + e);
	}
    }
    
    public Cell createCell (Integer p_ID, SystemModel p_sys, CellProgram p_cp, Vector args) 
    {
	CellFactory s = (CellFactory) cellModels.get (cellModelName);
	
	if (s == null) {
	    try {
		Class.forName(cellModelName);
		s = (CellFactory) cellModels.get(cellModelName);
		
		if (s == null) {
		    Assert.badBranch ("Could not create cell class: " + cellModelName);
		}
	    } catch (ClassNotFoundException e) {
		Assert.badBranch ("Could not find cell class: " + cellModelName);
	    } catch (NoClassDefFoundError e) {
		Assert.badBranch ("Error loading cell class: " + cellModelName);
	    }
	}
	
	return s.create (p_ID, p_sys, p_cp, cellModelOptions, args);
    }
    
    public CellProgram createCellProgram (String modelName, String modelOptions, BufferedReader inp)
	throws GenericParseException
    {
	ParserFactory s = (ParserFactory) parsers.get (modelName);
	
	if (s == null) {
	    try  {
		Class.forName (modelName);
		s = (ParserFactory) parsers.get(modelName);
		if (s == null) {
		    Assert.badBranch ("Could not find parser: " + modelName);
		}
	    } catch (ClassNotFoundException e) {
		Assert.badBranch ("Could not find parser class: " + modelName);
	    } catch (NoClassDefFoundError e) {
		Assert.badBranch ("Error loading parser class: " + modelName);
	    }
	}
	
	Parser p = s.create(inp,modelOptions);

	try {
	    return p.getCellProgram();
	} catch (GenericParseException e) {
	    throw new GenericParseException(e.toString());
	}
    }
    
    public SystemModel createSystem (String modelName, String modelOptions)
    {
	SystemFactory s = (SystemFactory) sysModels.get (modelName);

	if (s == null) {
	    try {
		Class.forName(modelName);
		// By now the static block with no function would
		// have been executed if the System was found.
		// the System is expected to have put its factory
		// in the hashtable.
		s = (SystemFactory) sysModels.get(modelName);

		if (s == null) {
		    Assert.badBranch ("Could not create system.");
		}
	    } catch (ClassNotFoundException e) {
		Assert.badBranch ("Could not find system class: " + modelName);
	    } catch (NoClassDefFoundError e) {
		Assert.badBranch ("Error loading system class: " + modelName);
	    } catch (Exception e) {
		Assert.badBranch ("Unhandled Exception creating system: " + e);
	    }
	}

	return s.create (this, modelOptions);
    }
}
